@@ -1,5 +1,5 @@
 """
-Easy image file reading & writing. Supports PGM/PPM/PNM/PFM/PNG/JPG/TIFF/EXR/RAW.
+Easy image file reading & writing. Supports PGM/PPM/PNM/PFM/PNG/BMP/JPG/TIFF/EXR/RAW.
 
 Example:
   image, maxval = imgio.imread("foo.png")
@@ -10,5 +10,5 @@ https://github.com/toaarnio/imgio
 
 from .imgio import *
 
-__version__ = "0.7.3"
+__version__ = "0.9.1"
 __all__ = ["imread", "imwrite", "selftest"]
