@@ -1,7 +1,7 @@
 #!/usr/bin/python3 -B
 
 """
-Easy image file reading & writing. Supports PGM/PPM/PNM/PFM/PNG/JPG/TIFF/EXR/RAW.
+Easy image file reading & writing. Supports PGM/PPM/PNM/PFM/PNG/BMP/JPG/TIFF/EXR/RAW.
 
 Example:
   image, maxval = imgio.imread("foo.png")
@@ -32,8 +32,8 @@ except ImportError:
 #
 ######################################################################################
 
-RW_FORMATS = [".pnm", ".pgm", ".ppm", ".pfm", ".png", ".jpg", ".jpeg", ".tif", ".tiff", ".insp", ".raw"]
-RO_FORMATS = RW_FORMATS + [".exr"]
+RW_FORMATS = [".pnm", ".pgm", ".ppm", ".pfm", ".png", ".jpg", ".jpeg", ".tif", ".tiff", ".insp", ".npy", ".raw"]
+RO_FORMATS = RW_FORMATS + [".exr", ".bmp"]
 
 def imread(filespec, width=None, height=None, bpp=None, raw_header_size=None, verbose=False):
     """
@@ -55,6 +55,9 @@ def imread(filespec, width=None, height=None, bpp=None, raw_header_size=None, ve
         _enforce(isinstance(height, int) and height >= 1, "height must be an integer >= 1; was %s"%(repr(height)))
         frame, maxval = _reraise(lambda: _read_raw(filespec, width, height, bpp, raw_header_size, verbose=verbose))
         return frame, maxval
+    if filetype == ".npy":
+        frame, maxval = _reraise(lambda: _read_npy(filespec, verbose))
+        return frame, maxval
     if filetype == ".pfm":
         frame, scale = _reraise(lambda: pfm.read(filespec, verbose))
         return frame, scale
@@ -64,7 +67,7 @@ def imread(filespec, width=None, height=None, bpp=None, raw_header_size=None, ve
     if filetype in [".pnm", ".pgm", ".ppm"]:
         frame, maxval = _reraise(lambda: pnm.read(filespec, verbose))
         return frame, maxval
-    if filetype in [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
+    if filetype in [".png", ".bmp", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
         _print(verbose, "Reading file %s "%(filespec), end='')
         formatstr = "jpg" if filetype == ".insp" else filetype[1:]
         frame = _reraise(lambda: _imread.imread(filespec, formatstr=formatstr))
@@ -82,7 +85,7 @@ def imwrite(filespec, image, maxval=255, packed=False, verbose=False):
     """
     Writes the given image to the given file, returns nothing. Grayscale images
     are expected to be provided as NumPy arrays with shape H x W, color images
-    with shape H x W x 3. Metadata, alpha channels, etc. are not supported.
+    with shape H x W x C. Metadata, alpha channels, etc. are not supported.
     """
     ImageIOError.error_message_prefix = "Failed to write %s: "%(repr(filespec))
     _enforce(isinstance(filespec, str), "filespec must be a string, was %s (%s)."%(type(filespec), repr(filespec)))
@@ -93,30 +96,33 @@ def imwrite(filespec, image, maxval=255, packed=False, verbose=False):
     _enforce(isinstance(maxval, int) or image.dtype.char in "efd", "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
     _enforce(1 <= maxval <= 65535 or image.dtype.char in "efd", "maxval must be an integer in [1, 65535]; was %s."%(repr(maxval)))
     _enforce(isinstance(verbose, bool), "verbose must be True or False, was %s (%s)."%(type(verbose), repr(verbose)))
-    _disallow(image.ndim not in [2, 3], "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
-    _disallow(image.ndim == 3 and image.shape[2] != 3, "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
+    _disallow(image.ndim not in [2, 3], "image.shape must be (m, n) or (m, n, c); was %s."%(str(image.shape)))
     _disallow(maxval > 255 and image.dtype == np.uint8, "maxval (%d) and image.dtype (%s) are inconsistent."%(maxval, image.dtype))
     _disallow(maxval <= 255 and image.dtype == np.uint16, "maxval (%d) and image.dtype (%s) are inconsistent."%(maxval, image.dtype))
     filename = os.path.basename(filespec)            # "path/image.pgm" => "image.pgm"
     basename, extension = os.path.splitext(filename)  # "image.pgm" => ("image", ".pgm")
     _enforce(len(basename) > 0, "filename `%s` must have at least 1 character + extension."%(filename))
-    _enforce(extension.lower() in RW_FORMATS, "unrecognized file extension `%s`."%(extension))
+    _enforce(extension.lower() in RW_FORMATS, "unrecognized or unsupported file extension `%s`."%(extension))
     filetype = extension.lower()
     if filetype == ".raw":
         _enforce(packed is False, "packed Bayer RAW images are not yet supported.")
         _enforce(image.ndim == 2, "image.shape must be (m, n) for a Bayer RAW; was %s."%(str(image.shape)))
         _reraise(lambda: _write_raw(filespec, image, maxval, packed, verbose))
     elif filetype == ".pfm":
+        _disallow(image.ndim == 3 and image.shape[2] != 3, "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
         _enforce(maxval >= 0.0, "maxval (scale) must be non-negative; was %s."%(repr(maxval)))
         _reraise(lambda: pfm.write(filespec, image, maxval, little_endian=True, verbose=verbose))
+    elif filetype == ".npy":
+        _reraise(lambda: _write_npy(filespec, image, verbose))
     elif filetype in [".pnm", ".pgm", ".ppm"]:
         _reraise(lambda: pnm.write(filespec, image, maxval, verbose))
     elif filetype in [".png", ".tif", ".tiff", ".jpg", ".jpeg", ".insp"]:
+        _disallow(image.ndim == 3 and image.shape[2] != 3, "image.shape must be (m, n) or (m, n, 3); was %s."%(str(image.shape)))
         _disallow(filetype in [".jpg", ".jpeg"] and maxval != 255, "maxval must be 255 for a JPEG; was %d."%(maxval))
         _disallow(filetype == ".png" and maxval not in [255, 65535], "maxval must be 255 or 65535 for a PNG; was %d."%(maxval))
         _print(verbose, "Writing file %s "%(filespec), end='')
         formatstr = "jpg" if filetype == ".insp" else filetype[1:]
-        _reraise(lambda: _imread.imsave(filespec, image, formatstr=formatstr))
+        _reraise(lambda: _imread.imsave(filespec, image, formatstr=formatstr, opts={'jpeg:quality': 95}))
         h, w = image.shape[:2]
         c = image.shape[2] if image.ndim > 2 else 1
         _print(verbose, "(w=%d, h=%d, c=%d, maxval=%d)"%(w, h, c, maxval))
@@ -181,7 +187,7 @@ def _exif_rotate(img, filespec):
         img = np.rot90(img, rot90_ccw_steps)  # 0/90/180/270 CCW
     return img
 
-def _read_exr(filespec, verbose=True):
+def _read_exr(filespec, verbose=False):
     exr = pyexr.open(filespec)
     data = exr.get()
     maxval = np.max(data)
@@ -189,6 +195,22 @@ def _read_exr(filespec, verbose=True):
     _print(verbose, "(w=%d, h=%d, c=%d, %s)"%(exr.width, exr.height, len(exr.channels), data.dtype))
     return data, maxval
 
+def _read_npy(filespec, verbose=False):
+    data = np.load(filespec)
+    _enforce(data.ndim == 3, "NumPy file %s image has unsupported shape %s"%(filespec, str(data.shape)))
+    maxval = np.max(data)
+    h, w, ch = data.shape
+    _print(verbose, "Reading NumPy file %s "%(filespec), end='')
+    _print(verbose, "(w=%d, h=%d, c=%d, %s)"%(w, h, ch, data.dtype))
+    return data, maxval
+
+def _write_npy(filespec, image, verbose=False):
+    _enforce(image.ndim == 3, "image.shape must be (m, n, c) for .npy; was %s."%(str(image.shape)))
+    h, w, ch = image.shape
+    if verbose:
+        print("Writing file %s (w=%d, h=%d, c=%d, %s)"%(filespec, w, h, ch, image.dtype))
+    np.save(filespec, image)
+
 def _read_raw(filespec, width, height, bpp, header_size=None, verbose=False):
     # Warning: hardcoded endianness (x86)
     with open(filespec, "rb") as infile:
@@ -279,11 +301,10 @@ class _TestImgIo(unittest.TestCase):
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "nonexisting.ppm")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "nonexisting.pgm")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "nonexisting.pfm")
+        self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "nonexisting.bmp")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "invalidformat.pfm")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "invalidformat.jpg")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "invalidformat.ppm")
-        self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "invalidtype.bmp")
-        self.assertRaisesRegexp(ImageIOError, "^Failed to read", imread, "invalidtype.bmp")
         self.assertRaisesRegexp(ImageIOError, "^Failed to read.*verbose", imread, "validimage.ppm", verbose="foo")
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "invalidfilename", pixels8b, 255)
         self.assertRaisesRegexp(ImageIOError, "^Failed to write", imwrite, "invalidtype.bmp", pixels8b, 255)
@@ -403,6 +424,21 @@ class _TestImgIo(unittest.TestCase):
             self.assertTrue(np.allclose(result, pixels))
             os.remove(tempfile)
 
+    def test_npy(self):
+        for dt in ["float16", "float32"]:
+            for shape in [(1, 1, 2), (1, 1, 9), (7, 11, 3), (9, 13, 31), (123, 321, 3)]:
+                scale = 3.141
+                tempfile = "imgio.test.npy"
+                print("Testing NPY reading & writing in %s mode, shape=%s..."%(dt, repr(shape)))
+                pixels = np.random.random(shape)  # float64 pixels
+                pixels = pixels.astype(dt)  # convert to float16/32
+                imwrite(tempfile, pixels, maxval=scale, verbose=False)
+                result, resscale = imread(tempfile, verbose=False)
+                self.assertEqual(result.dtype, dt)
+                self.assertEqual(result.shape, shape)
+                self.assertTrue(np.allclose(result, pixels))
+                os.remove(tempfile)
+
     def test_exif(self):
         print("Testing EXIF orientation handling...")
         thispath = os.path.dirname(os.path.abspath(__file__))
